/*
 * Copyright (c) 1999 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are subject to the Sun Community Source License
 * Jini(TM) Technology Core Platform, v.1.0 (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of
 * the License at http://java.sun.com/products/jini. Software distributed
 * under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied.  See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * CopyrightVersion v1.0_Jini
 */
package com.sun.jini.thread;

import com.sun.jini.constants.TimeConstants;

/**
 * An abstract class for building tasks that retry on failure after a
 * timeout period.  This builds upon <code>TaskManager</code> for task
 * execution and <code>WakeupManager</code> for retry scheduling.  You
 * extend <code>RetryTask</code>, implementing <code>tryOnce</code> to
 * represent a single attempt at the task.  If <code>tryOnce</code>
 * returns <code>true</code>, the attempt was successful and the task
 * is complete.  If <code>tryOnce</code> returns <code>false</code>,
 * the task is scheduled for a future retry using <code>WakeupManager</code>.
 * <p>
 * The default retry times are defined by this class's implementation
 * of <code>retryTime</code></code>.  You can override this method to
 * change the retry times.
 * <p>
 * It is legal to reuse the same task again and again, but only after
 * the task is complete and <code>reset</code> has been called.
 * Inserting the same task multiple times before it has either been
 * cancelled or completed successfully will generate unpredictable
 * behavior.
 *
 * @see TaskManager
 * @see WakeupManager
 * @author Ken Arnold
 */
import com.sun.jini.thread.WakeupManager.Ticket;

public abstract class RetryTask implements TaskManager.Task, TimeConstants {
    private TaskManager	manager;	// the TaskManager for this task
    private RetryTime	retry;		// the retry object for this task
    private boolean	cancelled;	// have we been cancelled?
    private boolean	complete;	// have we completed successfully?
    private Ticket	ticket;		// the WakeupManager ticket
    private long	startTime;	// the time when we first needed a retry
    private int		attempt;	// the current attempt number

    /**
     * Our WakeupManager for retry scheduling.
     */
    private static final WakeupManager wakeup =
	new WakeupManager(new WakeupManager.ThreadDesc(null, true));

    /**
     * Default delay backoffs.  These are converted from intervals to
     * "time since start" by the static block below.
     *
     * @see #retryTime
     */
    private static final long[] delays = {
	 1 * SECONDS,
	 5 * SECONDS,
	10 * SECONDS,
	 1 * MINUTES,
	 1 * MINUTES,
	 5 * MINUTES,
    };

    /**
     * The final delay, which is repeated indefinitely.
     *
     * @see #retryTime
     */
    //!! Should be final, but compiler won't allow this blank final
    private static long LAST_DELAY;

    /*
     * Make the delays the amount of time since the start -- it
     * is easier to declare the intervals, but the elapsed time is
     * more *useful*.
     */
    static {
	for (int i = 1; i < delays.length; i++)
	    delays[i] += delays[i - 1];
	LAST_DELAY = delays[delays.length - 1];
    }

    /**
     * Debug output stream.
     */
    private static final java.io.PrintWriter DEBUG =
	ThreadDebug.getWriter("retryTask");


    /**
     * Create a new <code>RetryTask</code> that will be scheduled with
     * the given manager.
     */
    public RetryTask(TaskManager manager) {
	this.manager = manager;
	reset();
    }

    /**
     * Make a single attempt.  Return <code>true</code> if the attempt
     * was successful.  If the attempt is not successful, the task
     * will be scheduled for a future retry.
     */
    public abstract boolean tryOnce();

    /**
     * The <code>run</code> method used as a <code>TaskManagerTask</code>.
     * This invokes <code>tryOnce</code>.  If it is not successful, it
     * schedules the task for a future retry at the time it gets by
     * invoking <code>retryTime</code>.
     *
     * @see #tryOnce
     * @see #startTime
     */
    public void run() {
	synchronized (this) {		// avoid retry if cancelled
	    if (cancelled)		// if they cancelled
		return;			// do nothing
	}

	boolean success = tryOnce();

	synchronized (this) {
	    if (!success) {		// if at first we don't succeed ...
		attempt++;
		long at = retryTime();	// ... try, try again
		if (DEBUG != null)
		    DEBUG.println("retry in " + (at - startTime));
		if (retry == null)	// only create it if we need to
		    retry = new RetryTime();
		ticket = wakeup.schedule(at, retry);
	    } else {
		complete = true;
		notifyAll();		// see waitFor()
	    }
	}
    }

    /**
     * Return the next time at which we should make another attempt.
     * This is <em>not</em> an interval, but the actual time.
     * <p>
     * The implementation is free to do as it pleases with the policy
     * here.  The default implementation is to delay using intervals of
     * 1 second, 5 seconds, 10 seconds, 1 minute, and 1 minute between
     * attempts, and then retrying every five minutes forever.
     */
    public long retryTime() {
	long delay = (attempt < delays.length ? delays[attempt] : LAST_DELAY);
	long nextTime = startTime + delay;
	if (DEBUG != null)
	    DEBUG.println("delay of " + delay / SECONDS + " seconds");
	return nextTime;
    }

    /**
     * Return the system time of the start of the retries.  This is the
     * system time immediately after the first unsuccessful invocation
     * of <code>tryOnce</code>.  If <code>tryOnce</code> has not returned
     * unsuccessfully once for the task, returns zero.
     */
    public synchronized long startTime() {
	return startTime;
    }

    /**
     * Return the attempt number, starting with zero.
     */
    public synchronized int attempt() {
	return attempt;
    }

    /**
     * Cancel the retrying of the task.  This ensures that there will be
     * no further attempts to invoke <code>tryOnce</code>.  It will not
     * interfere with any ongoing invocation of <code>tryOnce</code>
     * unless a subclass overrides this to do so.  Any override of this
     * method should invoke <code>super.cancel()</code>.
     */
    public synchronized void cancel() {
	cancelled = true;
	if (ticket != null)
	    wakeup.cancel(ticket);
	notifyAll();		// see waitFor()
    }

    /**
     * Return <code>true</code> if <code>cancel</code> has been invoked.
     */
    public synchronized boolean cancelled() {
	return cancelled;
    }

    /**
     * Return <code>true</code> if <code>tryOnce</code> has returned
     * successfully.
     */
    public synchronized boolean complete() {
	return complete;
    }

    public synchronized boolean waitFor() throws InterruptedException {
	while (!cancelled && !complete)
	    wait();
	return complete;
    }

    /**
     * Reset values for a new use of this task.
     */
    public void reset() {
	cancel();		// remove from the wakeup queue
	startTime = System.currentTimeMillis();
	cancelled = false;
	complete = false;
	ticket = null;
	attempt = 0;
    }

    /**
     * This is the runnable class for the <code>WakeupManager</code>,
     * since we need different implementations of
     * <code>WakeupManager.run</code> and <code>TaskManager.run</code>.
     */
    private class RetryTime implements Runnable {
	/**
	 * Time to retry the task.
	 */
	public void run() {
	    synchronized (RetryTask.this) {
		ticket = null;
	    }
	    manager.add(RetryTask.this);
	}
    };
}
